/*
 * Amazon FreeRTOS MQTT Echo Demo V1.1.0
 * Copyright (C) 2017 Amazon.com, Inc. or its affiliates.  All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 * http://aws.amazon.com/freertos
 * http://www.FreeRTOS.org
 */


/**
 * @file light_button.c
 * @brief MQTT Light toggle button demo.
 *
 */

/* Standard includes. */
#include "string.h"
#include "stdio.h"

/* FreeRTOS includes. */
#include "FreeRTOS.h"
#include "task.h"
#include "event_groups.h"
#include "semphr.h"

/* MQTT includes. */
#include "aws_mqtt_agent.h"

/* Credentials includes. */
#include "aws_clientcredential.h"

/**
 * @brief Switch between initial button handling (events) and improved handling (counting semaphore).
 */
#define USE_SOLUTION    0

/**
 * @brief Demo MQTT client ID.
 *
 * It must be unique per light button, e.g. through its S/N.
 */
#define configCLIENT_ID          ( ( const uint8_t * ) "Light_Button-12345678" )

/**
 * @brief The topic that the MQTT client publishes to.
 */
#define configTOPIC_NAME         ( ( const uint8_t * ) "freertos/demos/button" )

/**
 * @brief Dimension of the character array buffer used to hold data that is 
 * published to the MQTT broker (in the cloud).
 * 
 * Current payload is "{ 'button' : { 'id' : 'configCLIENT_ID' } }"
 */
#define configMAX_PUBLISH_DATA_LENGTH    50


#define configMAIN_TASK_STACK_SIZE                  ( configMINIMAL_STACK_SIZE * 3 )
#define configMAIN_TASK_PRIORITY                    ( tskIDLE_PRIORITY )

/* Timeout used when establishing a connection, which required TLS
 * negotiation. */
#define configMQTT_ECHO_TLS_NEGOTIATION_TIMEOUT          pdMS_TO_TICKS( 12000 )

/* Timeout used when performing MQTT operations that do not need extra time
 * to perform a TLS negotiation. */
#define configMQTT_TIMEOUT                               pdMS_TO_TICKS( 2500 )

/*-----------------------------------------------------------*/

/**
 * @brief Main task of our system.
 *
 * Waits for task event set by button press and publishes message.
 *
 * @param[in] pvParameters Parameters passed while creating the task. Unused in our
 * case.
 */
static void prvMainTask( void * pvParameters );

/**
 * @brief Creates an MQTT client and then connects to the MQTT broker.
 *
 * The MQTT broker end point is set by clientcredentialMQTT_BROKER_ENDPOINT.
 *
 * @return pdPASS if everything is successful, pdFAIL otherwise.
 */
static BaseType_t prvCreateClientAndConnectToBroker( void );

/**
 * @brief Publishes the message to the echoTOPIC_NAME topic.
 *
 * This is called every time the button is pressed.
 */
static void prvPublishMessage( void );

/**
 * @ brief The handle of the MQTT client object used by the system.
 */
static MQTTAgentHandle_t xMQTTHandle = NULL;

#if USE_SOLUTION == 0
/**
 * @ brief The event group to notify task on button interrupt.
 */
static EventGroupHandle_t xEventHandle;
#else
/**
  * @ brief The counting semaphore to count pending button presses.
  */
static SemaphoreHandle_t xCountHandle;
#endif


/*-----------------------------------------------------------*/

static BaseType_t prvCreateClientAndConnectToBroker( void )
{
    MQTTAgentReturnCode_t xReturned;
    BaseType_t xReturn = pdFAIL;
    MQTTAgentConnectParams_t xConnectParameters =
    {
        clientcredentialMQTT_BROKER_ENDPOINT, /* The URL of the MQTT broker to connect to. */
        pdFALSE,                              /* Set to pdTRUE if the provided URL is an IP address, otherwise set to pdFALSE. */
        clientcredentialMQTT_BROKER_PORT,     /* Port number on which the MQTT broker is listening. */
        configCLIENT_ID,                        /* Client Identifier of the MQTT client. It should be unique per broker. */
        0,                                    /* The length of the client Id, filled in later as not const. */
        pdTRUE,                               /* Set to pdTRUE to use TLS, pdFALSE to not use TLS. */
        NULL,                                 /* User data supplied to the callback. Can be NULL. */
        NULL,                                 /* Callback used to report various events. Can be NULL. */
        NULL,                                 /* Certificate used for secure connection. Can be NULL. */
        0                                     /* Size of certificate used for secure connection. */
    };

    /* Check this function has not already been executed. */
    configASSERT( xMQTTHandle == NULL );

    /* The MQTT client object must be created before it can be used.  The
     * maximum number of MQTT client objects that can exist simultaneously
     * is set by mqttconfigMAX_BROKERS. */
    xReturned = MQTT_AGENT_Create( &xMQTTHandle );

    if( xReturned == eMQTTAgentSuccess )
    {
        /* Fill in the MQTTAgentConnectParams_t member that is not const,
         * and therefore could not be set in the initializer (where
         * xConnectParameters is declared in this function). */
        xConnectParameters.usClientIdLength = ( uint16_t ) strlen( ( const char * ) configCLIENT_ID );

        /* Connect to the broker. */
        //configPRINTF( ( "Attempting to connect to %s.\r\n", clientcredentialMQTT_BROKER_ENDPOINT ) );
        configPRINTF( ( "Attempting to connect to ***.amazonaws.com.\r\n" ) );
        xReturned = MQTT_AGENT_Connect( xMQTTHandle,
                                        &xConnectParameters,
                                        configMQTT_ECHO_TLS_NEGOTIATION_TIMEOUT );

        if( xReturned != eMQTTAgentSuccess )
        {
            /* Could not connect, so delete the MQTT client. */
            ( void ) MQTT_AGENT_Delete( xMQTTHandle );
            configPRINTF( ( "ERROR: Failed to connect.\r\n" ) );
        }
        else
        {
            configPRINTF( ( "Connected.\r\n" ) );
            xReturn = pdPASS;
        }
    }

    return xReturn;
}
/*-----------------------------------------------------------*/

static void prvPublishMessage( void )
{
    MQTTAgentPublishParams_t xPublishParameters;
    MQTTAgentReturnCode_t xReturned;
    char cDataBuffer[ configMAX_PUBLISH_DATA_LENGTH ];

    /* Check this function is not being called before the MQTT client object has
     * been created. */
    configASSERT( xMQTTHandle != NULL );

    /* Create the message that will be published. */
    ( void ) snprintf( cDataBuffer, configMAX_PUBLISH_DATA_LENGTH, "{ \"button\" : { \"id\" : \"%s\" } }", configCLIENT_ID);

    /* Setup the publish parameters. */
    memset( &( xPublishParameters ), 0x00, sizeof( xPublishParameters ) );
    xPublishParameters.pucTopic = configTOPIC_NAME;
    xPublishParameters.pvData = cDataBuffer;
    xPublishParameters.usTopicLength = ( uint16_t ) strlen( ( const char * ) configTOPIC_NAME );
    xPublishParameters.ulDataLength = ( uint32_t ) strlen( cDataBuffer );
    xPublishParameters.xQoS = eMQTTQoS1;

    /* Publish the message. */
    xReturned = MQTT_AGENT_Publish( xMQTTHandle,
                                    &( xPublishParameters ),
                                    configMQTT_TIMEOUT );

    if( xReturned == eMQTTAgentSuccess )
    {
        configPRINTF( ( "Successfully published '%s'\r\n", cDataBuffer ) );
    }
    else
    {
        configPRINTF( ( "ERROR: Failed to publish '%s'\r\n", cDataBuffer ) );
    }

    /* Remove compiler warnings in case configPRINTF() is not defined. */
    ( void ) xReturned;
}
/*-----------------------------------------------------------*/

static void prvMainTask( void * pvParameters )
{
    BaseType_t        xReturned = pdPASS;
    uint32_t          ulNotified;
    EventBits_t       xBits;
    const TickType_t  xWait = pdMS_TO_TICKS( 30000UL );

    /* Avoid compiler warnings about unused parameters. */
    ( void ) pvParameters;

#if USE_SOLUTION == 0
    xEventHandle = xEventGroupCreate();
    configASSERT( xEventHandle != NULL );
#else
    xCountHandle = xSemaphoreCreateCounting(256U, 0U);
    configASSERT( xCountHandle != NULL );
#endif

    /* Create the MQTT client object and connect it to the MQTT broker. */
    xReturned = prvCreateClientAndConnectToBroker();

    if( xReturned == pdPASS )
    {
#if USE_SOLUTION == 0
        for ( ; ; )
        {
            /* Wait for event to be set from button interrupt. */
            xBits = xEventGroupWaitBits(xEventHandle, (1 << 0), pdTRUE, 0, xWait);
            /* Publish message to toggle the connected lights. */
            if (xBits & (1 << 0)) {
                prvPublishMessage();
            }
        }
#else
        for ( ; ; )
        {
            /*  */
            if (xSemaphoreTake(xCountHandle, xWait) == pdTRUE) {
            /* Publish message to toggle the connected lights. */
                prvPublishMessage();
            }
        }
#endif
    }
    else
    {
        configPRINTF( ( "Could not connect to broker.\r\n" ) );
    }

    /* Disconnect the client. */
    ( void ) MQTT_AGENT_Disconnect( xMQTTHandle, configMQTT_TIMEOUT );

    /* End the demo by deleting all created resources. */
    configPRINTF( ( "MQTT echo demo finished.\r\n" ) );
    vTaskDelete( NULL ); /* Delete this task. */
}

/*-----------------------------------------------------------*/

void HAL_GPIO_EXTI_Callback(uint16_t GPIO_Pin) {
  BaseType_t xHigherPriorityTaskWoken = pdFALSE;

  /* Record enter interrupt with SystemView instrumentation. */
#if defined(USE_SYSVIEW)
  SEGGER_SYSVIEW_RecordEnterISR();
#endif

#if USE_SOLUTION == 0
  /* Notify main task about button press via event. */
  if (xEventHandle) {
    xEventGroupSetBitsFromISR( xEventHandle, (1 << 0), &xHigherPriorityTaskWoken );
  }
#else
  if (xCountHandle) {
    xSemaphoreGiveFromISR( xCountHandle, &xHigherPriorityTaskWoken );
  }
#endif
  /* Record exit interrupt with SystemView instrumentation. */
#if defined(USE_SYSVIEW)
  if (xHigherPriorityTaskWoken == pdTRUE) {
    SEGGER_SYSVIEW_RecordExitISRToScheduler();
  } else {
    SEGGER_SYSVIEW_RecordExitISR();
  }
#endif
}

/*-----------------------------------------------------------*/

void vStartMQTTLightButton( void )
{
    configPRINTF( ( "Creating Button Main Task...\r\n" ) );
    ( void ) xTaskCreate( prvMainTask,                         /* The function that implements the task. */
                          "Main Task",                         /* The name to assign to the task being created. */
                          configMAIN_TASK_STACK_SIZE,          /* The size, in WORDS (not bytes), of the stack to allocate for the task being created. */
                          NULL,                                /* The task parameter is not being used. */
                          configMAIN_TASK_PRIORITY,            /* The priority at which the task being created will run. */
                          NULL );                              /* Not storing the task's handle. */
}
/*-----------------------------------------------------------*/
